from .brain_large import BrainLargeDataset
from .cortex import CortexDataset
from .dataset import GeneExpressionDataset
from .synthetic import SyntheticDataset
from .retina import RetinaDataset
from .cbmc import CbmcDataset
from .brain_small import BrainSmallDataset
from .hemato import HematoDataset
from .pbmc import PbmcDataset
from .loom import LoomDataset

__all__ = ['SyntheticDataset',
           'CortexDataset',
           'BrainLargeDataset',
           'RetinaDataset',
           'GeneExpressionDataset',
           'CbmcDataset',
           'BrainSmallDataset',
           'HematoDataset',
           'PbmcDataset',
           'LoomDataset']


def load_datasets(dataset_name, save_path='data/'):
    if dataset_name == 'synthetic':
        gene_dataset = SyntheticDataset()
    elif dataset_name == 'cortex':
        gene_dataset = CortexDataset()
    elif dataset_name == 'brain_large':
        gene_dataset = BrainLargeDataset(subsample_size=128, save_path=save_path)
    elif dataset_name == 'retina':
        gene_dataset = RetinaDataset(save_path=save_path)
    elif dataset_name == 'cbmc':
        gene_dataset = CbmcDataset(save_path=save_path)
    elif dataset_name == 'brain_small':
        gene_dataset = BrainSmallDataset(save_path=save_path)
    elif dataset_name == 'hemato':
        gene_dataset = HematoDataset(save_path=save_path)
    elif dataset_name == 'pbmc':
        gene_dataset = PbmcDataset(save_path=save_path)
    elif dataset_name[-5:] == ".loom":
        gene_dataset = LoomDataset(filename=dataset_name, save_path=save_path)
    else:
        raise "No such dataset available"
    return gene_dataset



# -*- coding: utf-8 -*-

"""Handling datasets.
For the moment, is initialized with a torch Tensor of size (n_cells, nb_genes)"""
import os

import numpy as np
import scipy.sparse as sp_sparse
import torch
import urllib.request
from sklearn.preprocessing import StandardScaler
from torch.utils.data import Dataset


class GeneExpressionDataset(Dataset):
    """Gene Expression dataset. It deals with:
    - log_variational expression -> torch.log(1 + X)
    - local library size normalization (mean, var) per batch
    """

    def __init__(self, X, local_means, local_vars, batch_indices, labels, gene_names=None):
        # Args:
        # Xs: a list of numpy tensors with .shape[1] identical (total_size*nb_genes)
        # or a list of scipy CSR sparse matrix,
        # or transposed CSC sparse matrix (the argument sparse must then be set to true)
        self.X = X
        self.total_size, self.nb_genes = self.X.shape
        self.local_means = local_means
        self.local_vars = local_vars
        self.labels = labels
        self.batch_indices = batch_indices
        self.n_batches = len(np.unique(batch_indices))
        self.dense = type(self.X) is np.ndarray
        self.n_labels = len(np.unique(labels))

        if gene_names is not None:
            self.gene_names = np.char.upper(gene_names)  # Take an upper case convention for gene names

    def __len__(self):
        return self.X.shape[0]

    def __getitem__(self, idx):
        return idx

    def download(self):
        r = urllib.request.urlopen(self.url)
        print("Downloading data")

        def readIter(f, blocksize=1000):
            """Given a file 'f', returns an iterator that returns bytes of
            size 'blocksize' from the file, using read()."""
            while True:
                data = f.read(blocksize)
                if not data:
                    break
                yield data

        # Create the path to save the data
        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)

        with open(self.save_path + self.download_name, 'wb') as f:
            for data in readIter(r):
                f.write(data)

    def download_and_preprocess(self):
        if not os.path.exists(self.save_path + self.download_name):
            self.download()
        return self.preprocess()

    def collate_fn(self, batch):
        indexes = np.array(batch)
        X = torch.FloatTensor(self.X[indexes]) if self.dense else torch.FloatTensor(self.X[indexes].toarray())
        return X, torch.FloatTensor(self.local_means[indexes]), \
            torch.FloatTensor(self.local_vars[indexes]), \
            torch.LongTensor(self.batch_indices[indexes]), \
            torch.LongTensor(self.labels[indexes])

    def subsample_genes(self, p_genes=1., subset_genes=None):
        n_cells, n_genes = self.X.shape
        new_n_genes = int(p_genes * n_genes) if type(p_genes) is not int else p_genes
        if subset_genes is None:
            print("Downsampling from %i to %i genes" % (n_genes, new_n_genes))
            std_scaler = StandardScaler(with_mean=False)
            std_scaler.fit(self.X.astype(np.float64))
            subset_genes = np.argsort(std_scaler.var_)[::-1][:new_n_genes]
        else:
            print("Downsampling from %i to %i genes" % (n_genes, len(subset_genes)))
        self.X = self.X[:, subset_genes]
        self.nb_genes = self.X.shape[1]
        self.gene_names = self.gene_names[subset_genes]

    def subsample_cells(self, p_cells=1.):
        n_cells, n_genes = self.X.shape
        new_n_cells = int(p_cells * n_genes) if type(p_cells) is not int else p_cells
        print("Downsampling from %i to %i cells" % (n_cells, new_n_cells))
        indices = np.argsort(self.X.sum(axis=1))[::-1][:new_n_cells]

        def _subsample(a, indices):
            return a[indices]

        self.X, self.local_means, self.local_vars, self.batch_indices, self.labels = (
            _subsample(a, indices) for a in (self.X, self.local_means, self.local_vars,
                                             self.batch_indices, self.labels))

    @staticmethod
    def get_attributes_from_matrix(X, batch_index=0, labels=None):
        log_counts = np.log(X.sum(axis=1))
        local_mean = (np.mean(log_counts) * np.ones((X.shape[0], 1))).astype(np.float32)
        local_var = (np.var(log_counts) * np.ones((X.shape[0], 1))).astype(np.float32)
        batch_index = batch_index * np.ones((X.shape[0], 1))
        labels = labels.reshape(-1, 1) if labels is not None else np.zeros_like(batch_index)
        return X, local_mean, local_var, batch_index, labels

    @staticmethod
    def get_attributes_from_list(Xs, list_labels=None):
        nb_genes = Xs[0].shape[1]
        assert all(X.shape[1] == nb_genes for X in Xs), "All tensors must have same size"

        new_Xs = []
        local_means = []
        local_vars = []
        batch_indices = []
        labels = []
        for i, X in enumerate(Xs):
            label = list_labels[i] if list_labels is not None else list_labels
            X, local_mean, local_var, batch_index, label = (
                GeneExpressionDataset.get_attributes_from_matrix(X, batch_index=i, labels=label)
            )
            new_Xs += [X]
            local_means += [local_mean]
            local_vars += [local_var]
            batch_indices += [batch_index]
            labels += [label]
        local_means = np.concatenate(local_means)
        local_vars = np.concatenate(local_vars)
        batch_indices = np.concatenate(batch_indices)
        labels = np.concatenate(labels)
        X = np.concatenate(new_Xs) if type(new_Xs[0]) is np.ndarray else sp_sparse.vstack(new_Xs)
        return X, local_means, local_vars, batch_indices, labels



from contextlib import ContextDecorator

import torch
import torch.nn as nn


class enable_grad(torch.enable_grad, ContextDecorator):
    pass


class no_grad(torch.no_grad, ContextDecorator):
    pass


class eval_modules:
    def __call__(self, function):
        def wrapper(*args, **kwargs):
            [a.eval() for a in args + tuple(kwargs.items()) if isinstance(a, nn.Module)]
            result = function(*args, **kwargs)
            [a.train() for a in args + tuple(kwargs.items()) if isinstance(a, nn.Module)]
            return result

        return wrapper


def to_cuda(tensors, async=True):
    return [t.cuda(async=async) for t in tensors]



import torch
import torch.nn as nn
from torch.distributions import Normal, Multinomial, kl_divergence as kl

from scvi.metrics.log_likelihood import log_nb_positive, log_zinb_positive
from scvi.models.modules import Classifier
from scvi.models.modules import Encoder, DecoderSCVI
from scvi.models.utils import broadcast_labels


# VAE model - for classification: VAEC
class VAEC(nn.Module):
    def __init__(self, n_input, n_labels, n_hidden=128, n_latent=10, n_layers=1, dropout_rate=0.1, dispersion="gene",
                 log_variational=True, reconstruction_loss="zinb", n_batch=0, y_prior=None, use_cuda=False):
        super(VAEC, self).__init__()
        self.dispersion = dispersion
        self.log_variational = log_variational
        self.reconstruction_loss = reconstruction_loss
        # Automatically desactivate if useless
        self.n_batch = 0 if n_batch == 1 else n_batch
        self.n_labels = 0 if n_labels == 1 else n_labels
        if self.n_labels == 0:
            raise ValueError("VAEC is only implemented for > 1 label dataset")

        if self.dispersion == "gene":
            self.px_r = torch.nn.Parameter(torch.randn(n_input, ))

        self.z_encoder = Encoder(n_input, n_hidden=n_hidden, n_latent=n_latent, n_layers=n_layers,
                                 dropout_rate=dropout_rate, n_cat=n_labels)
        self.l_encoder = Encoder(n_input, n_hidden=n_hidden, n_latent=1, n_layers=1,
                                 dropout_rate=dropout_rate)
        self.decoder = DecoderSCVI(n_latent, n_input, n_hidden=n_hidden, n_layers=n_layers,
                                   dropout_rate=dropout_rate, n_batch=n_batch, n_labels=n_labels)

        self.y_prior = y_prior if y_prior is not None else (1 / n_labels) * torch.ones(n_labels)
        self.classifier = Classifier(n_input, n_hidden, n_labels, n_layers=n_layers, dropout_rate=dropout_rate)

        self.use_cuda = use_cuda and torch.cuda.is_available()
        if self.use_cuda:
            self.cuda()
            self.y_prior = self.y_prior.cuda()

    def classify(self, x):
        x = torch.log(1 + x)
        return self.classifier(x)

    def sample_from_posterior_z(self, x, y):
        x = torch.log(1 + x)
        # Here we compute as little as possible to have q(z|x)
        qz_m, qz_v, z = self.z_encoder(x, y)
        return z

    def sample_from_posterior_l(self, x):
        x = torch.log(1 + x)
        # Here we compute as little as possible to have q(z|x)
        ql_m, ql_v, library = self.l_encoder(x)
        return library

    def get_sample_scale(self, x, y=None, batch_index=None):
        x = torch.log(1 + x)
        z = self.sample_from_posterior_z(x, y)
        px = self.decoder.px_decoder(z, batch_index, y)
        px_scale = self.decoder.px_scale_decoder(px)
        return px_scale

    def get_sample_rate(self, x, y=None, batch_index=None):
        x = torch.log(1 + x)
        z = self.sample_from_posterior_z(x, y)
        library = self.sample_from_posterior_l(x)
        px = self.decoder.px_decoder(z, batch_index, y)
        return self.decoder.px_scale_decoder(px) * torch.exp(library)

    def forward(self, x, local_l_mean, local_l_var, batch_index=None, y=None):
        is_labelled = False if y is None else True

        # Prepare for sampling
        xs, ys = (x, y)

        # Enumerate choices of label
        ys, xs, batch_index, local_l_var, local_l_mean = (
            broadcast_labels(
                ys, xs, batch_index, local_l_var, local_l_mean, n_broadcast=self.n_labels
            )
        )

        xs_ = xs
        if self.log_variational:
            xs_ = torch.log(1 + xs_)

        # Sampling
        qz_m, qz_v, z = self.z_encoder(xs_, ys)
        ql_m, ql_v, library = self.l_encoder(xs_)

        if self.dispersion == "gene-cell":
            px_scale, self.px_r, px_rate, px_dropout = self.decoder(self.dispersion, z, library, batch_index, y=ys)
        elif self.dispersion == "gene":
            px_scale, px_rate, px_dropout = self.decoder(self.dispersion, z, library, batch_index, y=ys)

        # Reconstruction Loss
        if self.reconstruction_loss == 'zinb':
            reconst_loss = -log_zinb_positive(xs, px_rate, torch.exp(self.px_r), px_dropout)
        elif self.reconstruction_loss == 'nb':
            reconst_loss = -log_nb_positive(xs, px_rate, torch.exp(self.px_r))

        # KL Divergence
        mean = torch.zeros_like(qz_m)
        scale = torch.ones_like(qz_v)

        kl_divergence_z = kl(Normal(qz_m, torch.sqrt(qz_v)), Normal(mean, scale)).sum(dim=1)
        kl_divergence_l = kl(Normal(ql_m, torch.sqrt(ql_v)), Normal(local_l_mean, torch.sqrt(local_l_var))).sum(dim=1)
        kl_divergence = kl_divergence_z + kl_divergence_l

        if is_labelled:
            return reconst_loss, kl_divergence

        reconst_loss = reconst_loss.view(self.n_labels, -1)
        kl_divergence = kl_divergence.view(self.n_labels, -1)

        if self.log_variational:
            x_ = torch.log(1 + x)

        probs = self.classifier(x_)
        reconst_loss = (reconst_loss.t() * probs).sum(dim=1)
        kl_divergence = (kl_divergence.t() * probs).sum(dim=1)
        kl_divergence += kl(Multinomial(probs=probs), Multinomial(probs=self.y_prior))

        return reconst_loss, kl_divergence



import torch

from scvi.metrics.stats import Stats
from scvi.utils import to_cuda, enable_grad


@enable_grad()
def adapt_encoder(vae, dataloader, n_path=10, n_epochs=50, record_freq=5):
    parameters = list(vae.z_encoder.parameters()) + list(vae.l_encoder.parameters())
    z_encoder_state = vae.z_encoder.state_dict()
    l_encoder_state = vae.l_encoder.state_dict()

    optimizer = torch.optim.Adam(parameters, eps=0.01)
    # Getting access to the stats during training
    stats = Stats(n_epochs=n_epochs, record_freq=record_freq, names=['test'], verbose=False)
    stats.callback(vae, dataloader)
    best_ll = stats.history["LL_test"][0]

    # Training the model
    for i in range(n_path):
        # Re-initialize to create new path
        vae.z_encoder.load_state_dict(z_encoder_state)
        vae.l_encoder.load_state_dict(l_encoder_state)
        for epoch in range(n_epochs):
            for i_batch, tensors in enumerate(dataloader):
                if vae.use_cuda:
                    tensors = to_cuda(tensors)
                sample_batch, local_l_mean, local_l_var, batch_index, labels = tensors
                sample_batch = sample_batch.type(torch.float32)

                reconst_loss, _ = vae(sample_batch, local_l_mean, local_l_var, batch_index=batch_index, y=labels)
                train_loss = torch.mean(reconst_loss)

                optimizer.zero_grad()
                train_loss.backward()
                optimizer.step()

            stats.callback(vae, dataloader)
        best_ll = min(min(stats.history["LL_test"]), best_ll)
    return best_ll



import numpy as np
import torch

from scvi.utils import to_cuda, no_grad, eval_modules


@no_grad()
@eval_modules()
def get_latent(vae, data_loader):
    latent = []
    batch_indices = []
    labels = []
    for tensors in data_loader:
        if vae.use_cuda:
            tensors = to_cuda(tensors)
        sample_batch, local_l_mean, local_l_var, batch_index, label = tensors
        sample_batch = sample_batch.type(torch.float32)
        latent += [vae.sample_from_posterior_z(sample_batch, y=label)]
        batch_indices += [batch_index]
        labels += [label]
    return torch.cat(latent), torch.cat(batch_indices), torch.cat(labels)


# CLUSTERING METRICS
def entropy_batch_mixing(latent_space, batches, max_number=500):
    # latent space: numpy matrix of size (number_of_cells, latent_space_dimension)
    # with the encoding of the different inputs in the latent space
    # batches: numpy vector with the batch indices of the cells
    n_samples = len(latent_space)
    keep_idx = np.random.choice(np.arange(n_samples), size=min(len(latent_space), max_number), replace=False)
    latent_space, batches = latent_space[keep_idx], batches[keep_idx]

    def entropy(hist_data):
        n_batches = len(np.unique(hist_data))
        if n_batches > 2:
            raise ValueError("Should be only two clusters for this metric")
        frequency = np.mean(hist_data == 1)
        if frequency == 0 or frequency == 1:
            return 0
        return -frequency * np.log(frequency) - (1 - frequency) * np.log(1 - frequency)

    n_samples = latent_space.shape[0]
    distance = np.zeros((n_samples, n_samples))
    neighbors_graph = np.zeros((n_samples, n_samples))
    for i in range(n_samples):
        for j in range(i, n_samples):
            distance[i, j] = distance[j, i] = sum((latent_space[i] - latent_space[j]) ** 2)

    for i, d in enumerate(distance):
        neighbors_graph[i, d.argsort()[:51]] = 1
    kmatrix = neighbors_graph - np.identity(latent_space.shape[0])

    score = 0
    for t in range(50):
        indices = np.random.choice(np.arange(latent_space.shape[0]), size=100)
        score += np.mean([entropy(
            batches[kmatrix[indices].nonzero()[1][kmatrix[indices].nonzero()[0] == i]]
        ) for i in range(100)])
    return score / 50



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""Tests for `scvi` package."""
from scvi.benchmark import run_benchmarks
from scvi.benchmark import run_benchmarks_classification
from scvi.models import VAEC, VAE, SVAEC


def test_synthetic_1():
    run_benchmarks("synthetic", n_epochs=1, use_batches=True, model=VAE)
    run_benchmarks_classification("synthetic", n_epochs=1, n_epochs_classifier=1)


def test_synthetic_2():
    run_benchmarks("synthetic", n_epochs=1, model=SVAEC, benchmark=True)


def test_cortex():
    run_benchmarks("cortex", n_epochs=1, model=VAEC)


def test_brain_large():
    run_benchmarks("brain_large", n_epochs=1, use_batches=False, tt_split=0.5, save_path='tests/data/')


def test_retina():
    run_benchmarks("retina", n_epochs=1, show_batch_mixing=False, save_path='tests/data/')


def test_cbmc():
    run_benchmarks("cbmc", n_epochs=1, show_batch_mixing=False, save_path='tests/data/')


def test_brain_small():
    run_benchmarks("brain_small", n_epochs=1, show_batch_mixing=False, save_path='tests/data/')


def test_hemato():
    run_benchmarks("hemato", n_epochs=1, show_batch_mixing=False, save_path='tests/data/HEMATO/')


def test_pbmc():
    run_benchmarks("pbmc", n_epochs=1, show_batch_mixing=False, save_path='tests/data/PBMC/')


def test_loom():
    run_benchmarks(dataset_name="retina.loom", n_epochs=1, show_batch_mixing=False, save_path='tests/data/')



import numpy as np

from . import GeneExpressionDataset


class SyntheticDataset(GeneExpressionDataset):
    def __init__(self, batch_size=200, nb_genes=100, n_batches=2, n_labels=3):
        # Generating samples according to a ZINB process
        data = np.random.negative_binomial(5, 0.3, size=(n_batches, batch_size, nb_genes))
        mask = np.random.binomial(n=1, p=0.7, size=(n_batches, batch_size, nb_genes))
        newdata = (data * mask)  # We put the batch index first
        labels = np.random.randint(0, n_labels, size=(n_batches, batch_size, 1))
        super(SyntheticDataset, self).__init__(
            *GeneExpressionDataset.get_attributes_from_list(newdata, list_labels=labels)
        )



import pandas as pd
import numpy as np

from .dataset import GeneExpressionDataset


class CbmcDataset(GeneExpressionDataset):
    url = "https://www.ncbi.nlm.nih.gov/geo/download/" + \
          "?acc=GSE100866&format=file&file=GSE100866%5FCBMC%5F8K%5F13AB%5F10X%2DRNA%5Fumi%2Ecsv%2Egz"

    def __init__(self, save_path='data/'):
        self.save_path = save_path

        # originally: GSE100866_CBMC_8K_13AB_10X-RNA_umi.csv.gz

        self.download_name = 'cbmc.csv.gz'

        expression_data, gene_names = self.download_and_preprocess()

        super(CbmcDataset, self).__init__(
            *GeneExpressionDataset.get_attributes_from_matrix(
                expression_data), gene_names=gene_names)

    def preprocess(self):
        print("Preprocessing cbmc data")
        expression = pd.read_csv(self.save_path + self.download_name, index_col=0, compression='gzip').T
        gene_names = np.array(expression.columns, dtype=str)

        selected = np.std(expression.values, axis=0).argsort()[-600:][::-1]
        expression_data = expression.values[:, selected]

        return expression_data, gene_names



import matplotlib.pyplot as plt
import numpy as np
import torch
from torch.utils.data import DataLoader
from torch.utils.data.sampler import SubsetRandomSampler

from scvi.dataset import CortexDataset, load_datasets
from scvi.metrics.adapt_encoder import adapt_encoder
from scvi.metrics.clustering import entropy_batch_mixing, get_latent
from scvi.metrics.differential_expression import get_statistics
from scvi.metrics.imputation import imputation
from scvi.metrics.visualization import show_t_sne
from scvi.models import VAE, SVAEC
from scvi.models.modules import Classifier
from scvi.train import train, train_classifier, train_semi_supervised


def run_benchmarks(dataset_name, model=VAE, n_epochs=1000, lr=1e-3, use_batches=False, use_cuda=True,
                   show_batch_mixing=True, benchmark=False, tt_split=0.9, save_path='data/'):
    # options:
    # - gene_dataset: a GeneExpressionDataset object
    # call each of the 4 benchmarks:
    # - log-likelihood
    # - imputation
    # - batch mixing
    # - cluster scores
    gene_dataset = load_datasets(dataset_name, save_path=save_path)
    example_indices = np.random.permutation(len(gene_dataset))
    tt_split = int(tt_split * len(gene_dataset))  # 90%/10% train/test split

    data_loader_train = DataLoader(gene_dataset, batch_size=128, pin_memory=use_cuda,
                                   sampler=SubsetRandomSampler(example_indices[:tt_split]),
                                   collate_fn=gene_dataset.collate_fn)
    data_loader_test = DataLoader(gene_dataset, batch_size=128, pin_memory=use_cuda,
                                  sampler=SubsetRandomSampler(example_indices[tt_split:]),
                                  collate_fn=gene_dataset.collate_fn)
    vae = model(gene_dataset.nb_genes, n_batch=gene_dataset.n_batches * use_batches, n_labels=gene_dataset.n_labels,
                use_cuda=use_cuda)
    stats = train(vae, data_loader_train, data_loader_test, n_epochs=n_epochs, lr=lr, benchmark=benchmark)

    if isinstance(vae, VAE):
        best_ll = adapt_encoder(vae, data_loader_test, n_path=1, n_epochs=1, record_freq=1)
        print("Best ll was :", best_ll)

    # - log-likelihood
    print("Log-likelihood Train:", stats.history["LL_train"][stats.best_index])
    print("Log-likelihood Test:", stats.history["LL_test"][stats.best_index])

    # - imputation
    imputation_test = imputation(vae, data_loader_test)
    print("Imputation score on test (MAE) is:", imputation_test.item())

    # - batch mixing
    if gene_dataset.n_batches == 2:
        latent, batch_indices, labels = get_latent(vae, data_loader_train)
        print("Entropy batch mixing :", entropy_batch_mixing(latent.cpu().numpy(), batch_indices.cpu().numpy()))
        if show_batch_mixing:
            show_t_sne(latent.cpu().numpy(), np.array([batch[0] for batch in batch_indices.cpu().numpy()]))

    # - differential expression
    if type(gene_dataset) == CortexDataset:
        get_statistics(vae, data_loader_train, M_sampling=1, M_permutation=1)  # 200 - 100000


# Pipeline to compare different semi supervised models
def run_benchmarks_classification(dataset_name, n_latent=10, n_epochs=10, n_epochs_classifier=10, lr=1e-2,
                                  use_batches=False, use_cuda=True, tt_split=0.9):
    gene_dataset = load_datasets(dataset_name)
    fig, axes = plt.subplots(1, 2, sharey=True, figsize=(12, 5))

    alpha = 100  # in Kingma, 0.1 * len(gene_dataset), but pb when : len(gene_dataset) >> 1

    # Create the dataset
    example_indices = np.random.permutation(len(gene_dataset))
    tt_split = int(tt_split * len(gene_dataset))  # 90%/10% train/test split

    data_loader_train = DataLoader(gene_dataset, batch_size=128, pin_memory=use_cuda,
                                   sampler=SubsetRandomSampler(example_indices[:tt_split]),
                                   collate_fn=gene_dataset.collate_fn)
    data_loader_test = DataLoader(gene_dataset, batch_size=128, pin_memory=use_cuda,
                                  sampler=SubsetRandomSampler(example_indices[tt_split:]),
                                  collate_fn=gene_dataset.collate_fn)

    # Now we try out the different models and compare the classification accuracy

    # ========== The M1 model ===========
    print("Trying M1 model")
    vae = VAE(gene_dataset.nb_genes, n_latent=n_latent,
              n_batch=gene_dataset.n_batches * use_batches, use_cuda=use_cuda,
              n_labels=gene_dataset.n_labels)
    train_semi_supervised(vae, data_loader_train, data_loader_test, n_epochs=n_epochs, lr=lr)

    # Then we train a classifier on the latent space
    cls = Classifier(n_input=n_latent, n_labels=gene_dataset.n_labels, n_layers=3, use_cuda=use_cuda)
    for param in vae.z_encoder.parameters():
        param.requires_grad = False
    cls_stats = train_classifier(vae, cls, data_loader_train, data_loader_test, n_epochs=n_epochs_classifier, lr=lr)

    axes[0].plot(cls_stats.history["Accuracy_train"], label='classifier')
    axes[1].plot(cls_stats.history["Accuracy_test"])

    # ========== The M1+M2 model, first encoder frozen ===========
    print("Trying out standard M1+M2 model")
    prior = torch.FloatTensor([(gene_dataset.labels == i).mean() for i in range(gene_dataset.n_labels)])

    svaec = SVAEC(gene_dataset.nb_genes, n_batch=gene_dataset.n_batches * use_batches, n_labels=gene_dataset.n_labels,
                  y_prior=prior, n_latent=n_latent, use_cuda=use_cuda)

    # Use a pretrained z encoder, freeze its weights
    svaec.z_encoder.load_state_dict(vae.z_encoder.state_dict())
    for param in svaec.z_encoder.parameters():
        param.requires_grad = False
    stats = train_semi_supervised(svaec, data_loader_train, data_loader_test, n_epochs=n_epochs, lr=lr,
                                  classification_ratio=alpha)

    # We don't train the first z encoder in this procedure
    axes[0].plot(stats.history["Accuracy_train"], label='M1+M2 (frozen)')
    axes[1].plot(stats.history["Accuracy_test"])

    # ========== The M1+M2 model trained jointly ===========
    print("Trying out M1+M2 optimized jointly")
    svaec = SVAEC(gene_dataset.nb_genes, n_labels=gene_dataset.n_labels, y_prior=prior, n_latent=n_latent,
                  use_cuda=use_cuda)

    stats = train_semi_supervised(svaec, data_loader_train, data_loader_test, n_epochs=n_epochs, lr=lr,
                                  classification_ratio=alpha)

    axes[0].plot(stats.history["Accuracy_train"], label='M1+M2 (train all)')
    axes[1].plot(stats.history["Accuracy_test"])

    # ========== Classifier trained on the latent space of M1+M2 ===========
    print("Trying to classify on M1+M2's z1 latent space")
    cls = Classifier(n_input=n_latent, n_labels=gene_dataset.n_labels, n_layers=3, use_cuda=use_cuda)

    stats = train_classifier(svaec, cls, data_loader_train, data_loader_test, n_epochs=n_epochs_classifier, lr=lr)

    axes[0].plot(stats.history["Accuracy_train"], label='M1+M2+classifier')
    axes[1].plot(stats.history["Accuracy_test"])

    # Now plot the results
    axes[0].set_ylim(0, 1)
    axes[0].set_ylabel('accuracy')
    axes[0].set_xlabel('n_epochs')
    axes[0].set_title('acc. train')
    axes[0].legend()
    axes[1].set_xlabel('n_epochs')
    axes[1].set_title('acc. test')

    plt.tight_layout()
    plt.savefig("result_classification.png")



import torch
import torch.nn as nn
from torch.distributions import Normal, Multinomial, kl_divergence as kl

from scvi.metrics.log_likelihood import log_zinb_positive
from scvi.models.modules import Decoder, Encoder, Classifier, DecoderSCVI
from scvi.models.utils import broadcast_labels


class SVAEC(nn.Module):
    '''
    "Stacked" variational autoencoder for classification - SVAEC
    (from the stacked generative model M1 + M2)
    '''

    def __init__(self, n_input, n_labels, n_hidden=128, n_latent=10, n_layers=1, dropout_rate=0.1, n_batch=0,
                 y_prior=None, use_cuda=False):
        super(SVAEC, self).__init__()
        self.n_labels = n_labels
        self.n_input = n_input

        self.y_prior = y_prior if y_prior is not None else (1 / self.n_labels) * torch.ones(self.n_labels)
        # Automatically desactivate if useless
        self.n_batch = 0 if n_batch == 1 else n_batch
        self.z_encoder = Encoder(n_input, n_hidden=n_hidden, n_latent=n_latent, n_layers=n_layers,
                                 dropout_rate=dropout_rate)
        self.l_encoder = Encoder(n_input, n_hidden=n_hidden, n_latent=1, n_layers=1,
                                 dropout_rate=dropout_rate)
        self.decoder = DecoderSCVI(n_latent, n_input, n_hidden=n_hidden, n_layers=n_layers,
                                   dropout_rate=dropout_rate, n_batch=n_batch)

        self.dispersion = 'gene'
        self.px_r = torch.nn.Parameter(torch.randn(n_input, ))

        # Classifier takes n_latent as input
        self.classifier = Classifier(n_latent, n_hidden, self.n_labels, n_layers, dropout_rate)
        self.encoder_z2_z1 = Encoder(n_input=n_latent, n_cat=self.n_labels, n_latent=n_latent, n_layers=n_layers)
        self.decoder_z1_z2 = Decoder(n_latent, n_latent, n_cat=self.n_labels, n_layers=n_layers)

        self.use_cuda = use_cuda and torch.cuda.is_available()
        if self.use_cuda:
            self.cuda()
            self.y_prior = self.y_prior.cuda()

    def classify(self, x):
        x_ = torch.log(1 + x)
        qz_m, _, z = self.z_encoder(x_)

        if self.training:
            return self.classifier(z)
        else:
            return self.classifier(qz_m)

    def sample_from_posterior_z(self, x, y=None):
        x = torch.log(1 + x)
        # Here we compute as little as possible to have q(z|x)
        qz_m, qz_v, z = self.z_encoder(x)
        return z

    def sample_from_posterior_l(self, x):
        x = torch.log(1 + x)
        # Here we compute as little as possible to have q(z|x)
        ql_m, ql_v, library = self.l_encoder(x)
        return library

    def get_sample_scale(self, x, y=None, batch_index=None):
        x = torch.log(1 + x)
        z = self.sample_from_posterior_z(x, y)
        px = self.decoder.px_decoder(z, batch_index, y)
        px_scale = self.decoder.px_scale_decoder(px)
        return px_scale

    def get_sample_rate(self, x, y=None, batch_index=None):
        x = torch.log(1 + x)
        z = self.sample_from_posterior_z(x)
        library = self.sample_from_posterior_l(x)
        px = self.decoder.px_decoder(z, batch_index, y)
        return self.decoder.px_scale_decoder(px) * torch.exp(library)

    def forward(self, x, local_l_mean, local_l_var, batch_index=None, y=None):
        is_labelled = False if y is None else True

        xs, ys = (x, y)
        xs_ = torch.log(1 + xs)
        qz1_m, qz1_v, z1_ = self.z_encoder(xs_)
        z1 = z1_
        # Enumerate choices of label
        ys, xs, batch_index, local_l_var, local_l_mean, qz1_m, qz1_v, z1 = (
            broadcast_labels(
                ys, xs, batch_index, local_l_var, local_l_mean, qz1_m, qz1_v, z1, n_broadcast=self.n_labels
            )
        )

        xs_ = torch.log(1 + xs)

        qz2_m, qz2_v, z2 = self.encoder_z2_z1(z1, ys)
        pz1_m, pz1_v = self.decoder_z1_z2(z2, ys)

        # Sampling
        ql_m, ql_v, library = self.l_encoder(xs_)  # let's keep that ind. of y

        px_scale, px_rate, px_dropout = self.decoder(self.dispersion, z1, library, batch_index)

        reconst_loss = -log_zinb_positive(xs, px_rate, torch.exp(self.px_r), px_dropout)

        # KL Divergence
        mean = torch.zeros_like(qz2_m)
        scale = torch.ones_like(qz2_v)

        kl_divergence_z2 = kl(Normal(qz2_m, torch.sqrt(qz2_v)), Normal(mean, scale)).sum(dim=1)
        loss_z1 = (- Normal(pz1_m, torch.sqrt(pz1_v)).log_prob(z1) +
                   Normal(qz1_m, torch.sqrt(qz1_v)).log_prob(z1)).sum(dim=1)
        kl_divergence_l = kl(Normal(ql_m, torch.sqrt(ql_v)), Normal(local_l_mean, torch.sqrt(local_l_var))).sum(dim=1)
        kl_divergence = kl_divergence_z2 + loss_z1 + kl_divergence_l

        if is_labelled:
            return reconst_loss, kl_divergence

        reconst_loss = reconst_loss.view(self.n_labels, -1)
        kl_divergence = kl_divergence.view(self.n_labels, -1)

        probs = self.classifier(z1_)
        reconst_loss = (reconst_loss.t() * probs).sum(dim=1)
        kl_divergence = (kl_divergence.t() * probs).sum(dim=1)

        kl_divergence += kl(Multinomial(probs=probs), Multinomial(probs=self.y_prior))

        return reconst_loss, kl_divergence



#!/usr/bin/env python

"""Run all the benchmarks with specific parameters"""
import argparse

from scvi.benchmark import run_benchmarks

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--epochs", type=int, default=250, help="how many times to process the dataset")
    parser.add_argument("--dataset", type=str, default="cortex", help="which dataset to process")
    parser.add_argument("--nobatches", action='store_true', help="whether to ignore batches")
    parser.add_argument("--nocuda", action='store_true',
                        help="whether to use cuda (will apply only if cuda is available")
    parser.add_argument("--benchmark", action='store_true',
                        help="whether to use cuda (will apply only if cuda is available")
    args = parser.parse_args()

    run_benchmarks(args.dataset, n_epochs=args.epochs, use_batches=(not args.nobatches), use_cuda=(not args.nocuda),
                   show_batch_mixing=True, benchmark=args.benchmark)



import numpy as np
import torch

from scvi.utils import to_cuda, no_grad, eval_modules


@no_grad()
@eval_modules()
def imputation(vae, data_loader, rate=0.1):
    distance_list = torch.FloatTensor([])
    for tensorlist in data_loader:
        if vae.use_cuda:
            tensorlist = to_cuda(tensorlist)
        sample_batch, local_l_mean, local_l_var, batch_index, labels = tensorlist
        sample_batch = sample_batch.type(torch.float32)
        dropout_batch = sample_batch.clone()
        indices = torch.nonzero(dropout_batch)
        i, j = indices[:, 0], indices[:, 1]
        ix = torch.LongTensor(np.random.choice(range(len(i)), int(np.floor(rate * len(i))), replace=False))
        dropout_batch[i[ix], j[ix]] *= 0

        if vae.use_cuda:
            ix, i, j = to_cuda([ix, i, j], async=False)
        px_rate = vae.get_sample_rate(dropout_batch, labels, batch_index=batch_index)
        distance_list = torch.cat([distance_list, torch.abs(px_rate[i[ix], j[ix]] - sample_batch[i[ix], j[ix]]).cpu()])
    return torch.median(distance_list)



"""File for computing log likelihood of the data"""

import torch
import torch.nn.functional as F

from scvi.utils import to_cuda, no_grad, eval_modules


@no_grad()
@eval_modules()
def compute_log_likelihood(vae, data_loader):
    # Iterate once over the data_loader and computes the total log_likelihood
    log_lkl = 0
    for i_batch, tensors in enumerate(data_loader):
        if vae.use_cuda:
            tensors = to_cuda(tensors)
        sample_batch, local_l_mean, local_l_var, batch_index, labels = tensors
        sample_batch = sample_batch.type(torch.float32)
        reconst_loss, kl_divergence = vae(sample_batch, local_l_mean, local_l_var, batch_index=batch_index,
                                          y=labels)
        log_lkl += torch.sum(reconst_loss).item()
    n_samples = (len(data_loader.dataset)
                 if not (hasattr(data_loader, 'sampler') and hasattr(data_loader.sampler, 'indices')) else
                 len(data_loader.sampler.indices))
    return log_lkl / n_samples


def log_zinb_positive(x, mu, theta, pi, eps=1e-8):
    """
    Note: All inputs are torch Tensors
    log likelihood (scalar) of a minibatch according to a zinb model.
    Notes:
    We parametrize the bernoulli using the logits, hence the softplus functions appearing

    Variables:
    mu: mean of the negative binomial (has to be positive support) (shape: minibatch x genes)
    theta: inverse dispersion parameter (has to be positive support) (shape: minibatch x genes)
    pi: logit of the dropout parameter (real support) (shape: minibatch x genes)
    eps: numerical stability constant
    """

    # theta is the dispersion rate. If .ndimension() == 1, it is shared for all cells (regardless of batch or labels)
    if theta.ndimension() == 1:
        theta = theta.view(1, theta.size(0))  # In this case, we reshape theta for broadcasting

    case_zero = (F.softplus((- pi + theta * torch.log(theta + eps) - theta * torch.log(theta + mu + eps)))
                 - F.softplus(-pi))

    case_non_zero = - pi - F.softplus(-pi) + theta * torch.log(theta + eps) - theta * torch.log(
        theta + mu + eps) + x * torch.log(mu + eps) - x * torch.log(theta + mu + eps) + torch.lgamma(
        x + theta + eps) - torch.lgamma(theta + eps) - torch.lgamma(x + 1)

    res = torch.mul((x < eps).type(torch.float32), case_zero) + torch.mul((x > eps).type(torch.float32), case_non_zero)
    return torch.sum(res, dim=-1)


def log_nb_positive(x, mu, theta, eps=1e-8):
    """
    Note: All inputs should be torch Tensors
    log likelihood (scalar) of a minibatch according to a nb model.

    Variables:
    mu: mean of the negative binomial (has to be positive support) (shape: minibatch x genes)
    theta: inverse dispersion parameter (has to be positive support) (shape: minibatch x genes)
    eps: numerical stability constant
    """
    res = theta * torch.log(theta + eps) - theta * torch.log(theta + mu + eps) + x * torch.log(
        mu + eps) - x * torch.log(theta + mu + eps) + torch.lgamma(x + theta) - torch.lgamma(
        theta.view(1, theta.size(0))) - torch.lgamma(
        x + 1)
    return torch.sum(res)



from itertools import cycle

import torch
from torch.nn import functional as F

from scvi.metrics.stats import Stats, EarlyStopping
from scvi.utils import to_cuda, enable_grad


@enable_grad()
def train(vae, data_loader_train, data_loader_test, n_epochs=20, lr=0.001, kl=None, benchmark=False):
    # Defining the optimizer
    optimizer = torch.optim.Adam(filter(lambda p: p.requires_grad, vae.parameters()), lr=lr)

    # Getting access to the stats during training
    stats = Stats(n_epochs=n_epochs, benchmark=benchmark)
    stats.callback(vae, data_loader_train, data_loader_test)
    early_stopping = EarlyStopping(benchmark=benchmark)

    # Training the model
    for epoch in range(n_epochs):
        total_train_loss = 0
        for i_batch, (tensors_train) in enumerate(data_loader_train):
            if vae.use_cuda:
                tensors_train = to_cuda(tensors_train)
            sample_batch_train, local_l_mean_train, local_l_var_train, batch_index_train, labels_train = tensors_train
            sample_batch_train = sample_batch_train.type(torch.float32)

            if kl is None:
                kl_ponderation = min(1, epoch / 400.)
            else:
                kl_ponderation = kl

            reconst_loss_train, kl_divergence_train = vae(sample_batch_train, local_l_mean_train, local_l_var_train,
                                                          batch_index=batch_index_train, y=labels_train)
            train_loss = torch.mean(reconst_loss_train + kl_ponderation * kl_divergence_train)

            batch_size = sample_batch_train.size(0)
            total_train_loss += train_loss.item() * batch_size
            optimizer.zero_grad()
            train_loss.backward()
            optimizer.step()

        if not early_stopping.update(total_train_loss):
            break
        stats.callback(vae, data_loader_train, data_loader_test)
    stats.display_time()
    stats.set_best_params(vae)
    return stats


@enable_grad()
def train_semi_supervised(vae, data_loader_train, data_loader_test, n_epochs=20, lr=0.001, kl=None, benchmark=False,
                          classification_ratio=0):
    # Defining the optimizer
    optimizer = torch.optim.Adam(filter(lambda p: p.requires_grad, vae.parameters()), lr=lr)

    # Getting access to the stats during training
    stats = Stats(n_epochs=n_epochs, benchmark=benchmark)
    stats.callback(vae, data_loader_train, data_loader_test)
    early_stopping = EarlyStopping(benchmark=benchmark)

    # Training the model
    for epoch in range(n_epochs):
        # initialize kl, reconst
        total_train_loss = 0
        for i_batch, (tensors_train, tensors_test) in enumerate(zip(data_loader_train, cycle(data_loader_test))):
            with torch.no_grad():
                if vae.use_cuda:
                    tensors_train = to_cuda(tensors_train)
                    tensors_test = to_cuda(tensors_test)
            sample_batch_train, local_l_mean_train, local_l_var_train, batch_index_train, labels_train = tensors_train
            sample_batch_test, local_l_mean_test, local_l_var_test, batch_index_test, _ = tensors_test
            sample_batch_train = sample_batch_train.type(torch.float32)
            sample_batch_test = sample_batch_test.type(torch.float32)

            if kl is None:
                kl_ponderation = min(1, epoch / 400.)
            else:
                kl_ponderation = kl

            reconst_loss_train, kl_divergence_train = vae(sample_batch_train, local_l_mean_train, local_l_var_train,
                                                          batch_index=batch_index_train, y=labels_train)
            reconst_loss_test, kl_divergence_test = vae(sample_batch_test, local_l_mean_test, local_l_var_test,
                                                        batch_index=batch_index_test, y=None)

            train_loss = torch.mean(reconst_loss_train + kl_ponderation * kl_divergence_train)
            test_loss = torch.mean(reconst_loss_test + kl_ponderation * kl_divergence_test)
            train_test_loss = train_loss + test_loss

            # Add a classification loss (most semi supervised VAE papers do)
            if hasattr(vae, "classify") and labels_train is not None:
                classification_loss = F.cross_entropy(vae.classify(sample_batch_train), labels_train.view(-1))
                train_test_loss += classification_loss * classification_ratio

            batch_size = sample_batch_train.size(0)
            total_train_loss += train_loss.item() * batch_size
            optimizer.zero_grad()
            train_test_loss.backward()
            optimizer.step()

        if not early_stopping.update(total_train_loss):
            break
        stats.callback(vae, data_loader_train, data_loader_test)
    stats.display_time()
    return stats


@enable_grad()
def train_classifier(vae, classifier, data_loader_train, data_loader_test, n_epochs=250, lr=0.1):
    # Train classifier on the z latent space of a vae
    optimizer = torch.optim.Adam(classifier.parameters(), lr=lr, eps=0.01)

    # Getting access to the stats during training
    stats = Stats(n_epochs=n_epochs)
    stats.callback(vae, data_loader_train, data_loader_test, classifier=classifier)

    for epoch in range(1, n_epochs + 1):
        for i_batch, tensors in enumerate(data_loader_train):
            if vae.use_cuda:
                tensors = to_cuda(tensors)
            sample_batch_train, _, _, _, labels_train = tensors
            sample_batch_train = sample_batch_train.type(torch.float32)
            # Get the latent space
            z = vae.sample_from_posterior_z(sample_batch_train, labels_train)

            optimizer.zero_grad()
            loss = F.cross_entropy(classifier(z), labels_train.view(-1))
            loss.backward()
            optimizer.step()

        stats.callback(vae, data_loader_train, data_loader_test, classifier=classifier)

    return stats



from .dataset import GeneExpressionDataset
import loompy
import numpy as np
import scipy.sparse as sp_sparse


class LoomDataset(GeneExpressionDataset):

    def __init__(self, filename, save_path):
        self.filename = filename
        self.save_path = save_path

        data, labels, cell_batches = self.preprocess()

        # Currently only supports retina.loom dataset
        data_with_info = np.hstack((data, labels, cell_batches))
        first_batch = data_with_info[data_with_info[:, -1] == 0.0]
        second_batch = data_with_info[data_with_info[:, -1] == 1.0]
        first_batch = first_batch[:, :-1]
        second_batch = second_batch[:, :-1]
        print("Finished preprocessing dataset")
        super(LoomDataset, self).__init__(*GeneExpressionDataset.get_attributes_from_list(
            [sp_sparse.csr_matrix(first_batch[:, :-1]), sp_sparse.csr_matrix(second_batch[:, :-1])],
            list_labels=[first_batch[:, -1], second_batch[:, -1]]))

    def preprocess(self):
        ds = loompy.connect(self.save_path + self.filename)
        cell_batches = ds.ca['Batch_id']
        cell_batches = np.reshape(cell_batches, (cell_batches.shape[0], 1))

        labels = np.array(ds.ca['Clusters'])
        labels = np.reshape(labels, (labels.shape[0], 1))

        data = np.array(ds[:, :]).T  # change matrix to cells by genes
        ds.close()

        return data, labels, cell_batches



import pandas as pd
import numpy as np
import urllib.request
import os
from scipy import io
from pathlib import Path
from zipfile import ZipFile


from .dataset import GeneExpressionDataset


class PbmcDataset(GeneExpressionDataset):
    datazip_url = "https://github.com/romain-lopez/scVI-reproducibility/raw/master/additional/data.zip"

    def __init__(self, save_path='data/PBMC/'):
        self.save_path = save_path

        self.download_name = "gene_info.csv"
        self.gene_names_filename = "michael_gene_names.csv"  # from romain, manually put into PBMC folder
        self.count_filename = "count.mtx"  # from romain, manually put into PBMC folder

        expression_data, gene_names = self.download_and_preprocess()

        super(PbmcDataset, self).__init__(
            *GeneExpressionDataset.get_attributes_from_matrix(
                expression_data), gene_names=gene_names)

    def download(self):
        r = urllib.request.urlopen(self.datazip_url)
        print("Downloading data.zip")

        def readIter(f, blocksize=1000):
            while True:
                data = f.read(blocksize)
                if not data:
                    break
                yield data

        directory = str(Path(self.save_path).parent) + '/'

        if not os.path.exists(directory):
            os.makedirs(directory)
        with open(directory + 'data.zip', 'wb') as f:
            for data in readIter(r):
                f.write(data)

        with ZipFile(directory + 'data.zip', 'r') as zip:
            zip.extractall(path=directory)

    def preprocess(self):
        print("Preprocessing pbmc data")

        gene_names = np.array(pd.read_csv(self.save_path + self.gene_names_filename, index_col=0)["x"], dtype=str)
        expression = pd.DataFrame(io.mmread(self.save_path + self.count_filename).T.A, columns=gene_names)

        micro_array_result = pd.read_csv(self.save_path + self.download_name)
        expression = expression[micro_array_result["ENSG"]]

        de_expression = np.copy(expression.values)
        de_gene_names = np.array(expression.columns, dtype=str)

        return de_expression, de_gene_names



import collections

import torch
from torch import nn as nn
from torch.distributions import Normal

from scvi.models.utils import one_hot


class FCLayers(nn.Module):
    def __init__(self, n_in, n_out, n_hidden=128, n_layers=1, dropout_rate=0.1):
        super(FCLayers, self).__init__()
        layers_dim = [n_in] + (n_layers - 1) * [n_hidden] + [n_out]
        self.fc_layers = FCLayers._sequential(layers_dim, dropout_rate=dropout_rate)

    def forward(self, x, *os):
        return self.fc_layers(x)

    @staticmethod
    def create(n_in, n_out, n_cat=0, n_hidden=128, n_layers=1, dropout_rate=0.1):
        if type(n_cat) is int:
            if n_cat == 0:
                return FCLayers(n_in, n_out, n_hidden=n_hidden, n_layers=n_layers, dropout_rate=dropout_rate)
            else:
                return OneHotFCLayers(n_in, n_out, n_cat=n_cat, n_hidden=n_hidden, n_layers=n_layers,
                                      dropout_rate=dropout_rate)
        elif type(n_cat) is list:
            return ManyOneHotFCLayers(n_in, n_out, n_cat_list=n_cat,
                                      n_hidden=n_hidden, n_layers=n_layers, dropout_rate=dropout_rate)

    @staticmethod
    def _sequential(layers_dim, n_cat=0, dropout_rate=0.1):
        return nn.Sequential(collections.OrderedDict(
            [('Layer {}'.format(i), nn.Sequential(
                nn.Dropout(p=dropout_rate),
                nn.Linear(n_in + n_cat, n_out),
                nn.BatchNorm1d(n_out, eps=1e-3, momentum=0.99),
                nn.ReLU())) for i, (n_in, n_out) in enumerate(zip(layers_dim[:-1], layers_dim[1:]))]))


class OneHotFCLayers(nn.Module):
    def __init__(self, n_in, n_out, n_cat, n_hidden=128, n_layers=1, dropout_rate=0.1):
        super(OneHotFCLayers, self).__init__()
        layers_dim = [n_in] + (n_layers - 1) * [n_hidden] + [n_out]
        self.n_cat = n_cat
        self.fc_layers = FCLayers._sequential(layers_dim, n_cat=n_cat, dropout_rate=dropout_rate)

    def forward(self, x, o, *os):
        if o.size(1) != self.n_cat:
            o = one_hot(o, self.n_cat)
        for layer in self.fc_layers:
            x = layer(torch.cat((x, o), 1))
        return x


class ManyOneHotFCLayers(nn.Module):
    def __init__(self, n_in, n_out, n_cat_list, n_hidden=128, n_layers=1, dropout_rate=0.1):
        super(ManyOneHotFCLayers, self).__init__()
        layers_dim = [n_in] + (n_layers - 1) * [n_hidden] + [n_out]
        self.n_cat_list = n_cat_list
        self.fc_layers = FCLayers._sequential(layers_dim, n_cat=sum(n_cat_list), dropout_rate=dropout_rate)

    def forward(self, x, *os):
        one_hot_os = []
        for i, o in enumerate(os):
            if o is not None and self.n_cat_list[i]:
                one_hot_o = o
                if o.size(1) != self.n_cat_list[i]:
                    one_hot_o = one_hot(o, self.n_cat_list[i])
                elif o.size(1) == 1 and self.n_cat_list[i] == 1:
                    one_hot_o = o.type(torch.float32)
                one_hot_os += [one_hot_o]
        for layer in self.fc_layers:
            x = layer(torch.cat((x,) + tuple(one_hot_os), 1))
        return x


# Encoder
class Encoder(nn.Module):
    def __init__(self, n_input, n_hidden=128, n_latent=10, n_cat=0, n_layers=1, dropout_rate=0.1):
        super(Encoder, self).__init__()
        self.encoder = FCLayers.create(n_in=n_input, n_out=n_hidden, n_cat=n_cat, n_layers=n_layers, n_hidden=n_hidden,
                                       dropout_rate=dropout_rate)
        self.mean_encoder = nn.Linear(n_hidden, n_latent)
        self.var_encoder = nn.Linear(n_hidden, n_latent)

    def reparameterize(self, mu, var):
        return Normal(mu, var.sqrt()).rsample()

    def forward(self, x, o=None):
        # Parameters for latent distribution
        q = self.encoder(x, o)
        q_m = self.mean_encoder(q)
        q_v = torch.exp(torch.clamp(self.var_encoder(q), -5, 5))  # (computational stability safeguard)
        latent = self.reparameterize(q_m, q_v)
        return q_m, q_v, latent


# Decoder
class DecoderSCVI(nn.Module):
    def __init__(self, n_latent, n_input, n_hidden=128, n_layers=1, dropout_rate=0.1, n_batch=0, n_labels=0):
        super(DecoderSCVI, self).__init__()
        self.n_batch = n_batch
        self.px_decoder = FCLayers.create(n_in=n_latent, n_out=n_hidden, n_layers=n_layers, n_hidden=n_hidden,
                                          dropout_rate=dropout_rate, n_cat=[n_batch, n_labels])

        # mean gamma
        self.px_scale_decoder = nn.Sequential(nn.Linear(n_hidden, n_input), nn.Softmax(dim=-1))

        # dispersion: here we only deal with gene-cell dispersion case
        self.px_r_decoder = nn.Linear(n_hidden, n_input)

        # dropout
        self.px_dropout_decoder = nn.Linear(n_hidden, n_input)

    def forward(self, dispersion, z, library, batch_index=None, y=None):
        # The decoder returns values for the parameters of the ZINB distribution
        px = self.px_decoder(z, batch_index, y)
        px_scale = self.px_scale_decoder(px)
        px_dropout = self.px_dropout_decoder(px)
        # Clamp to high value: exp(12) ~ 160000 to avoid nans (computational stability)
        px_rate = torch.exp(torch.clamp(library, max=12)) * px_scale
        if dispersion == "gene-cell":
            px_r = self.px_r_decoder(px)
            return px_scale, px_r, px_rate, px_dropout
        else:  # dispersion == "gene" / "gene-batch" / "gene-label"
            return px_scale, px_rate, px_dropout


# Decoder
class Decoder(nn.Module):
    def __init__(self, n_latent, n_output, n_cat=0, n_hidden=128, n_layers=1, dropout_rate=0.1):
        super(Decoder, self).__init__()
        self.decoder = FCLayers.create(n_in=n_latent, n_out=n_hidden, n_cat=n_cat, n_layers=n_layers,
                                       n_hidden=n_hidden, dropout_rate=dropout_rate)

        self.mean_decoder = nn.Linear(n_hidden, n_output)
        self.var_decoder = nn.Linear(n_hidden, n_output)

    def forward(self, x, o=None):
        # Parameters for latent distribution
        p = self.decoder(x, o)
        p_m = self.mean_decoder(p)
        p_v = torch.exp(self.var_decoder(p))
        return p_m, p_v


# Classifier
class Classifier(nn.Module):
    def __init__(self, n_input, n_hidden=128, n_labels=10, n_layers=1, dropout_rate=0.1, use_cuda=False):
        super(Classifier, self).__init__()
        self.layers = FCLayers(n_in=n_input, n_out=n_hidden, n_layers=n_layers, n_hidden=n_hidden,
                               dropout_rate=dropout_rate)
        self.classifier = nn.Sequential(self.layers, nn.Linear(n_hidden, n_labels), nn.Softmax(dim=-1))

        self.use_cuda = use_cuda and torch.cuda.is_available()
        if self.use_cuda:
            self.cuda()

    def forward(self, x):
        # Parameters for latent distribution
        return self.classifier(x)



import pandas as pd
import os
from zipfile import ZipFile
import urllib.request
from pathlib import Path
from .dataset import GeneExpressionDataset


class HematoDataset(GeneExpressionDataset):
    url = "https://www.ncbi.nlm.nih.gov/geo/download/?acc=GSM2388072&format=file&" + \
          "file=GSM2388072%5Fbasal%5Fbone%5Fmarrow%2Eraw%5Fumifm%5Fcounts%2Ecsv%2Egz"
    datazip_url = "https://github.com/romain-lopez/scVI-reproducibility/raw/master/additional/data.zip"

    def __init__(self, save_path='data/HEMATO/'):
        self.save_path = save_path

        # Originally: GSM2388072_basal_bone_marrow.raw_umifm_counts.csv.gz

        self.download_name = "bBM.raw_umifm_counts.csv.gz"
        self.gene_names_filename = "bBM.filtered_gene_list.paper.txt"
        self.spring_and_pba_filename = "bBM.spring_and_pba.csv"

        if not os.path.isfile(str(Path(save_path).parent) + '/data.zip'):
            self.download_datazip()

        expression_data, gene_names = self.download_and_preprocess()

        super(HematoDataset, self).__init__(
            *GeneExpressionDataset.get_attributes_from_matrix(
                expression_data), gene_names=gene_names)

    def download_datazip(self):
        r = urllib.request.urlopen(self.datazip_url)
        print("Downloading data.zip")

        def readIter(f, blocksize=1000):
            while True:
                data = f.read(blocksize)
                if not data:
                    break
                yield data

        directory = str(Path(self.save_path).parent) + '/'

        if not os.path.exists(directory):
            os.makedirs(directory)
        with open(directory + 'data.zip', 'wb') as f:
            for data in readIter(r):
                f.write(data)

        with ZipFile(directory + 'data.zip', 'r') as zip:
            zip.extractall(path=directory)

    def preprocess(self):
        print("Preprocessing Hemato data")

        raw_counts = pd.read_csv(self.save_path + self.download_name, compression='gzip')
        raw_counts.drop(raw_counts.index[raw_counts["library_id"] == "basal_bm1"], inplace=True)

        spring_and_pba = pd.read_csv(self.save_path + self.spring_and_pba_filename)
        with open(self.save_path + self.gene_names_filename) as f:
            gene_filter_list = f.read()

        gene_names = gene_filter_list.splitlines()

        data = raw_counts.merge(spring_and_pba, how="inner")
        expression_data = data[gene_names]

        expression_data = expression_data.values

        return expression_data, gene_names



import matplotlib.pyplot as plt
import numpy as np
from sklearn.manifold import TSNE

plt.switch_backend('agg')


def show_t_sne(latent, labels, n_samples=1000):
    idx_t_sne = np.random.permutation(len(latent))[:n_samples]
    if latent.shape[1] != 2:
        latent = TSNE().fit_transform(latent[idx_t_sne])
    plt.figure(figsize=(10, 10))
    plt.scatter(latent[:, 0], latent[:, 1], c=(np.array(labels)[idx_t_sne]).ravel(), edgecolors='none')
    plt.axis("off")
    plt.tight_layout()
    print("saving tsne figure as tsne.png")
    plt.savefig("tsne.png")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

package_name = 'scvi'

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('requirements.txt') as requires:
    requirements = [l.strip() for l in requires]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

author = 'Romain Lopez, Jeffrey Regier, Maxime Langevin, Edouard Mehlman, Yining Liu'

setup(
    author=author,
    author_email='romain_lopez@berkeley.edu',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
    description="Single-cell Variational Inference",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='scvi',
    name=package_name,
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/YosefLab/scVI',
    version='0.1.1',
    zip_safe=False,
)



import collections
import time

import numpy as np
from scipy.sparse import csc_matrix
import h5py
from sklearn.preprocessing import StandardScaler

from .dataset import GeneExpressionDataset

GeneBCMatrix = collections.namedtuple('GeneBCMatrix', ['gene_ids', 'gene_names', 'barcodes', 'matrix'])

batch_idx_10x = [1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0,
                 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0,
                 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0,
                 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]


def subsample_barcodes(gbm, barcode_indices, unit_test=False):
    barcodes = gbm.barcodes[barcode_indices] if not unit_test else gbm.barcodes
    return GeneBCMatrix(gbm.gene_ids, gbm.gene_names, barcodes,
                        gbm.matrix[:, barcode_indices])


def subsample_genes(gbm, genes_indices, unit_test=False):
    gene_ids = gbm.gene_ids[genes_indices] if not unit_test else gbm.gene_ids
    gene_names = gbm.gene_names[genes_indices] if not unit_test else gbm.gene_names
    return GeneBCMatrix(gene_ids, gene_names, gbm.barcodes,
                        gbm.matrix[genes_indices, :])


def get_expression(gbm, gene_name):
    gene_indices = np.where(gbm.gene_names == gene_name)[0]
    if len(gene_indices) == 0:
        raise Exception("%s was not found in list of gene names." % gene_name)
    return gbm.matrix[gene_indices[0], :].toarray().squeeze()


class BrainLargeDataset(GeneExpressionDataset):
    url = "http://cf.10xgenomics.com/samples/cell-exp/1.3.0/1M_neurons/1M_neurons_filtered_gene_bc_matrices_h5.h5"

    def __init__(self, subsample_size=None, save_path='data/', nb_genes_kept=720):
        """
        :param subsample_size: In thousands of barcodes kept (by default 1*1000=1000 kept)
        :param unit_test: A boolean to indicate if we use pytest subsampled file
        """
        self.subsample_size = subsample_size
        self.save_path = save_path
        self.nb_genes_kept = nb_genes_kept
        # originally: "1M_neurons_filtered_gene_bc_matrices_h5.h5"

        self.download_name = "genomics.h5"

        Xs = self.download_and_preprocess()
        super(BrainLargeDataset, self).__init__(
            *GeneExpressionDataset.get_attributes_from_list(Xs)
        )

    def preprocess(self):
        print("Preprocessing Brain Large data")
        tic = time.time()

        filtered_matrix_h5 = self.save_path + self.download_name
        with h5py.File(filtered_matrix_h5) as f:
            dset = f["mm10"]
            matrix = csc_matrix((dset['data'], dset['indices'], dset['indptr']), shape=dset['shape'])
            gene_bc_matrix = GeneBCMatrix(dset['genes'][:], dset['gene_names'][:], dset['barcodes'][:], matrix)

        # Downsample *barcodes* from 1306127 to 100000 (~1/10) to
        matrix = gene_bc_matrix.matrix[:, :100000]
        std_scaler = StandardScaler(with_mean=False)
        std_scaler.fit(matrix.transpose().astype(np.float64))
        subset_genes = np.argsort(std_scaler.var_)[::-1][:self.nb_genes_kept]
        subsampled_matrix = subsample_genes(gene_bc_matrix, subset_genes, unit_test=(self.save_path == 'tests/data/'))
        print("%d genes subsampled" % subsampled_matrix.matrix.shape[0])
        print("%d cells subsampled" % subsampled_matrix.matrix.shape[1])

        batch_id = np.random.randint(0, 2, size=subsampled_matrix.matrix.T.shape[0])

        # Choose if Xs should be sparse
        if self.subsample_size is not None:
            idx = np.random.permutation(len(batch_id))[:self.subsample_size]
            X = subsampled_matrix.matrix.T[idx]
            b = batch_id[idx]
            Xs = [X[b == batch].A for batch in (0, 1)]
        else:
            X = subsampled_matrix.matrix.T
            b = batch_id
            Xs = [X[b == batch] for batch in (0, 1)]

        toc = time.time()
        print("Preprocessing finished in : %d sec." % int(toc - tic))

        return Xs  # the list of X (n_barcodes*n_genes)



import csv

import numpy as np

from .dataset import GeneExpressionDataset


class CortexDataset(GeneExpressionDataset):
    url = "https://storage.googleapis.com/linnarsson-lab-www-blobs/blobs/cortex/expression_mRNA_17-Aug-2014.txt"

    def __init__(self):
        # Generating samples according to a ZINB process
        self.save_path = 'data/'
        self.download_name = 'expression.bin'
        expression_data, labels, gene_names = self.download_and_preprocess()

        super(CortexDataset, self).__init__(
            *GeneExpressionDataset.get_attributes_from_matrix(
                expression_data,
                labels=labels),
            gene_names=gene_names)

    def preprocess(self):
        print("Preprocessing Cortex data")
        rows = []
        gene_names = []
        with open(self.save_path + self.download_name, 'r') as csvfile:
            data_reader = csv.reader(csvfile, delimiter='\t')
            clusters = None
            for i, row in enumerate(data_reader):
                if i == 8:  # 7 + 1 in pandas
                    clusters = np.array(row, dtype=str)[2:]
                if i >= 11:  # 10 + 1 in pandas
                    rows.append(row[1:])
                    gene_names.append(row[0])

        cell_types, labels = np.unique(clusters, return_inverse=True)

        expression_data = np.array(rows, dtype=np.int).T[1:]
        gene_names = np.array(gene_names, dtype=np.str)

        selected = np.std(expression_data, axis=0).argsort()[-558:][::-1]
        expression_data = expression_data[:, selected]
        gene_names = gene_names[selected]

        return expression_data, labels, gene_names



import torch

from scvi.utils import no_grad, eval_modules, to_cuda


@no_grad()
@eval_modules()
def compute_accuracy(vae, data_loader, classifier=None):
    all_y_pred = []
    all_labels = []

    for i_batch, tensors in enumerate(data_loader):
        if vae.use_cuda:
            tensors = to_cuda(tensors)
        sample_batch, _, _, _, labels = tensors
        sample_batch = sample_batch.type(torch.float32)
        all_labels += [labels.view(-1)]

        if classifier is not None:
            # Then we use the specified classifier
            mu_z, _, _ = vae.z_encoder(sample_batch)
            y_pred = classifier(mu_z).argmax(dim=-1)
        else:
            # Then the vae must implement a classify function
            y_pred = vae.classify(sample_batch).argmax(dim=-1)
        all_y_pred += [y_pred]

    accuracy = (torch.cat(all_y_pred) == torch.cat(all_labels)).type(torch.float32).mean().item()

    return accuracy



import time
from collections import defaultdict

import numpy as np

from scvi.metrics.classification import compute_accuracy
from scvi.metrics.log_likelihood import compute_log_likelihood
from scvi.models import VAE, VAEC, SVAEC


class Stats:
    def __init__(self, verbose=True, record_freq=5, n_epochs=-1, benchmark=False, names=['train', 'test', 'val']):
        self.verbose = verbose
        self.record_freq = record_freq
        self.n_epochs = n_epochs
        self.benchmark = benchmark
        self.names = names

        self.best_params = None
        self.best_score = +np.inf
        self.begin = time.time()
        self.epoch = 0
        self.history = defaultdict(lambda: [])

    def callback(self, model, *data_loaders, classifier=None):
        if self.epoch == 0 or self.epoch == self.n_epochs or (
                        self.epoch % self.record_freq == 0 and not self.benchmark):
            # In this case we do add the stats
            if self.verbose:
                print("EPOCH [%d/%d]: " % (self.epoch, self.n_epochs))
            for data_loader, name in zip(data_loaders, self.names):
                self.add_ll(model, data_loader, name=name)
                self.add_accuracy(model, data_loader, classifier=classifier, name=name)
            self.save_best_params(model)
        self.epoch += 1

    def add_ll(self, model, data_loader, name='train'):
        models = [VAE, VAEC, SVAEC]
        if type(model) in models:
            log_likelihood = compute_log_likelihood(model, data_loader)
            self.history["LL_%s" % name].append(log_likelihood)
            if self.verbose:
                print("LL %s is: %4f" % (name, log_likelihood))

    def add_accuracy(self, model, data_loader, classifier=None, name='train'):
        models = [VAEC, SVAEC]
        if type(model) in models or classifier:
            accuracy = compute_accuracy(model, data_loader, classifier)
            self.history["Accuracy_%s" % name].append(accuracy)
            if self.verbose:
                print("Accuracy %s is: %4f" % (name, accuracy))

    def save_best_params(self, model):
        current_score = self.history["LL_%s" % self.names[0]][-1]
        if current_score < self.best_score:
            self.best_index = self.epoch // self.record_freq
            self.best_score = current_score
            self.best_params = model.state_dict()

    def set_best_params(self, model):
        model.load_state_dict(self.best_params)

    def display_time(self):
        end = time.time()
        print("Total runtime for " + str(self.epoch) + " epochs is: " + str((end - self.begin))
              + " seconds for a mean per epoch runtime of " + str((end - self.begin) / self.epoch) + " seconds.")


class EarlyStopping:
    def __init__(self, patience=250, threshold=0.01, benchmark=False):
        self.benchmark = benchmark
        self.patience = patience
        self.threshold = threshold
        self.current_performances = np.ones((patience))
        self.epoch = 0

    def update(self, scalar):
        self.epoch += 1
        if self.benchmark or self.epoch <= self.patience:
            return True
        else:
            # Shift
            self.current_performances[:-1] = self.current_performances[1:]
            self.current_performances[-1] = scalar

            # Compute improvement
            improvement = ((self.current_performances[-1] - self.current_performances[0])
                           / self.current_performances[0])

            # Returns true if improvement is good enough
            return improvement >= self.threshold



from .loom import LoomDataset
import os


class RetinaDataset(LoomDataset):
    url = 'https://github.com/YosefLab/scVI-data/raw/master/retina.loom'

    def __init__(self, save_path='data/'):

        self.save_path = save_path
        self.download_name = 'retina.loom'

        if not os.path.exists(self.save_path + self.download_name):
            self.download()

        print("Finished preprocessing Retina dataset")
        super(RetinaDataset, self).__init__(filename=self.download_name, save_path=self.save_path)



import numpy as np

import tarfile
from scipy import io

from .dataset import GeneExpressionDataset


class BrainSmallDataset(GeneExpressionDataset):
    url = "http://cf.10xgenomics.com/samples/cell-exp/2.1.0/neuron_9k/" + \
          "neuron_9k_filtered_gene_bc_matrices.tar.gz"

    def __init__(self, save_path='data/'):
        self.save_path = save_path

        fn = "filtered_gene_bc_matrices"
        self.download_name = fn + ".tar.gz"
        self.gene_filename = fn + "/mm10/genes.tsv"
        self.expression_filename = fn + "/mm10/matrix.mtx"

        expression_data, gene_names = self.download_and_preprocess()

        super(BrainSmallDataset, self).__init__(
            *GeneExpressionDataset.get_attributes_from_matrix(
                expression_data), gene_names=gene_names)

    def preprocess(self):
        print("Preprocessing Brain Small data")
        gene_names = []

        tar = tarfile.open(self.save_path + self.download_name)
        gene_file = tar.extractfile(tar.getmember(self.gene_filename))

        lines = gene_file.readlines()
        for line in lines:
            gene_names.append(line[:-2])  # remove "\n"

        expression_file = tar.extractfile(tar.getmember(self.expression_filename))
        expression_data = io.mmread(expression_file).T.A

        gene_names = np.array(gene_names, dtype=np.str)

        selected = np.std(expression_data, axis=0).argsort()[-3000:][::-1]
        expression_data = expression_data[:, selected]
        gene_names = gene_names[selected]

        return expression_data, gene_names
